/*
    This is part of pyahocorasick Python module.
    
    Windows declarations

    Author    : Wojciech Muła, wojciech_mula@poczta.onet.pl
    WWW       : http://0x80.pl
    License   : BSD-3-Clause (see LICENSE)
*/

#ifndef PYAHCORASICK_WINDOWS_H__
#define PYAHCORASICK_WINDOWS_H__

typedef unsigned char       uint8_t;
typedef short unsigned int  uint16_t;

#define PY_OBJECT_HEAD_INIT	PyVarObject_HEAD_INIT(NULL, 0)

#endif
